/****************************************************************************
* Together - Simple Videoconference System
* Copyright (C) 2010 Marius Gligor info@gmax.ws
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/
#include "BaseCapture.h"

//---------------------------------------------------------------------------
CBaseCapture::CBaseCapture()
{
    // graph
    pGraph = NULL;
    pBuilder = NULL;
    pControl = NULL;
    pEvent = NULL;
}
//---------------------------------------------------------------------------
CBaseCapture::~CBaseCapture()
{
    // graph
    SAFE_RELEASE(pControl);
    SAFE_RELEASE(pBuilder);
    SAFE_RELEASE(pGraph);
    SAFE_RELEASE(pEvent);
}
//---------------------------------------------------------------------------
// Filter Graph
BOOL CBaseCapture::Start()
{
    if(!pControl)
    {
        return FALSE;
    }
    hr = pControl->Run();
    return (hr == S_OK);
}
//---------------------------------------------------------------------------
BOOL CBaseCapture::Stop()
{
    if(!pControl)
    {
        return FALSE;
    }
    hr = pControl->Stop();
    return (hr == S_OK);
}
//---------------------------------------------------------------------------
BOOL CBaseCapture::FilterProperty(HWND hWnd, IBaseFilter **pFilter)
{
    CAUUID caGUID;
    FILTER_INFO FilterInfo;
    ISpecifyPropertyPages *pSpecify;

    // Discover if this filter contains a property page
    hr = (*pFilter)->QueryInterface(
        IID_ISpecifyPropertyPages,
        (void **)&pSpecify);
    if(hr != S_OK)
    {
        return FALSE;
    }

    hr = (*pFilter)->QueryFilterInfo(&FilterInfo);
    if (hr != S_OK)
    {
        return FALSE;
    }

    hr = pSpecify->GetPages(&caGUID);
    if (hr != S_OK)
    {
        return FALSE;
    }

    SAFE_RELEASE(pSpecify);

    // Display the filter's property page
    OleCreatePropertyFrame(
        hWnd,                   // Parent window
        0,                      // x (Reserved)
        0,                      // y (Reserved)
        FilterInfo.achName,     // Caption for the dialog box
        1,                      // Number of filters
        (IUnknown **)pFilter,   // Pointer to the filter
        caGUID.cElems,          // Number of property pages
        caGUID.pElems,          // Pointer to property page CLSIDs
        0,                      // Locale identifier
        0,                      // Reserved
        NULL                    // Reserved
    );

    CoTaskMemFree(caGUID.pElems);
    SAFE_RELEASE(FilterInfo.pGraph);

    return TRUE;
}
//---------------------------------------------------------------------------
BOOL CBaseCapture::AddRenderer(GUID guid, IBaseFilter **pFilter, LPTSTR Name)
{
    // Create Renderer
    hr = ::CoCreateInstance(guid,
        NULL,
        CLSCTX_INPROC_SERVER,
        IID_IBaseFilter,
        (void **)pFilter);
    if(hr != S_OK)
    {
        return FALSE;
    }
    // add filter
    hr = pGraph->AddFilter(*pFilter, Name);
    return (hr == S_OK);
}
//---------------------------------------------------------------------------
BOOL CBaseCapture::AddSource(REFCLSID Category,
    LPTSTR Device, IBaseFilter **pFilter, LPTSTR Name)
{
    BOOL Result = FALSE;
    // Create the System Device Enumerator.
    ICreateDevEnum *pSysDevEnum;
    hr = ::CoCreateInstance(CLSID_SystemDeviceEnum,
        NULL,
        CLSCTX_INPROC_SERVER,
        IID_ICreateDevEnum,
        (void **)&pSysDevEnum);
    if(hr != S_OK)
    {
        return FALSE;
    }
    // Obtain a class enumerator for the video compressor category.
    IEnumMoniker *pEnumCat;
    hr = pSysDevEnum->CreateClassEnumerator(Category, &pEnumCat, 0);
    if(hr != S_OK)
    {
        return FALSE;
    }
    // Enumerate the monikers.
    VARIANT varName;
    ULONG cFetched = 0;
    IPropertyBag *pPropBag;
    IMoniker *pMoniker;
    while(!Result && S_OK == pEnumCat->Next(1, &pMoniker, &cFetched))
    {
        hr = pMoniker->BindToStorage(0, 0, IID_IPropertyBag, (void **)&pPropBag);
        if(hr == S_OK)
        {
            // To retrieve the friendly name of the filter, do the following:
            VariantInit(&varName);
            hr = pPropBag->Read(L"FriendlyName", &varName, 0);
            if ((hr == S_OK) && (_tcscmp(Device, varName.bstrVal) == 0))
            {
                hr = pMoniker->BindToObject(0, 0, IID_IBaseFilter, (void**)pFilter);
                hr = pGraph->AddFilter(*pFilter, Name);
                Result = TRUE;
            }
            VariantClear(&varName);
            SAFE_RELEASE(pPropBag);
        }
        SAFE_RELEASE(pMoniker);
    }
    // clean up
    SAFE_RELEASE(pEnumCat);
    SAFE_RELEASE(pSysDevEnum);
    return Result;
}
//---------------------------------------------------------------------------
BOOL CBaseCapture::AddGrabber(IBaseFilter **pFilter,
    LPTSTR Name, ISampleGrabber **pGrabber)
{
    // Create the Sample Grabber.
    hr = ::CoCreateInstance(CLSID_SampleGrabber,
        NULL,
        CLSCTX_INPROC_SERVER,
        IID_IBaseFilter,
        (void**)pFilter);
    if(hr != S_OK)
    {
        return FALSE;
    }

    hr = pGraph->AddFilter(*pFilter, Name);
    if(hr != S_OK)
    {
        return FALSE;
    }

    // Settings
    hr = (*pFilter)->QueryInterface(IID_ISampleGrabber, (void**)pGrabber);
    return (hr == S_OK);
}
//---------------------------------------------------------------------------
void WINAPI CBaseCapture::DeleteMediaType(AM_MEDIA_TYPE *pmt)
{
    if(pmt)
    {
        FreeMediaType(*pmt);
        CoTaskMemFree((PVOID)pmt);
    }
}
//---------------------------------------------------------------------------
void WINAPI CBaseCapture::FreeMediaType(AM_MEDIA_TYPE& mt)
{
    if (mt.cbFormat != 0)
    {
        CoTaskMemFree((PVOID)mt.pbFormat);
        // Strictly unnecessary but tidier
        mt.cbFormat = 0;
        mt.pbFormat = NULL;
    }
    SAFE_RELEASE(mt.pUnk);
}
//---------------------------------------------------------------------------
LPCTSTR CBaseCapture::GetLastError()
{
    TCHAR msg[MAX_ERROR_TEXT_LEN] = _T("No error.");
    if(hr != S_OK)
    {
        DWORD len = AMGetErrorText(hr, msg, MAX_ERROR_TEXT_LEN);
        if(len)
        {
            msg[len - 2] = 0;
        }
    }
    return &msg[0];
}
//---------------------------------------------------------------------------

